#include "DataPacking.inc"
#include "TAA.inc"

// #define PHYSICALLY_BASED_MAX_ROUGHNESS // If disabled, materials that are 100% rough will have zero specular highlights (legacy SEUS behavior). If enabled, even 100% rough materials still show some subtle specular highlights. This can cause slowdown if "Full Raytrace Reflections" is on!
#define SUNLIGHT_LEAK_FIX // Fixes light leaking in caves. May cause issues with large outdoor overhanging structures!





float saturate(float x)
{
	return clamp(x, 0.0, 1.0);
}

vec3 saturate(vec3 x)
{
	return clamp(x, vec3(0.0), vec3(1.0));
}

vec2 saturate(vec2 x)
{
	return clamp(x, vec2(0.0), vec2(1.0));
}

// polynomial smooth min (k = 0.1); https://www.iquilezles.org/www/articles/smin/smin.htm
float SmoothMin( float a, float b, float k )
{
    float h = clamp( 0.5+0.5*(b-a)/k, 0.0, 1.0 );
    return mix( b, a, h ) - k*h*(1.0-h);
}

float SmoothMax( float a, float b, float k )
{
    return -SmoothMin(-a, -b, k);
}

vec3 nozero(vec3 x)
{
	if (abs(x.x) < 0.001)
		x.x = 0.001;

	if (abs(x.y) < 0.001)
		x.y = 0.001;

	if (abs(x.z) < 0.001)
		x.z = 0.001;

	return x;
}

vec3 Fract01(vec3 pos)
{
    #if 0
	vec3 posf = fract(pos);

	for (int i = 0; i < 3; i++)
	{
		if (posf[i] == 0.0)
		{
			posf[i] = 1.0;
		}
	}

	return posf;
    #else
    return fract(pos - vec3(0.00001)) + vec3(0.00001);
    #endif
}

vec3 ClipCheck(vec3 color)
{
	color.r = color.r > 1.0 ? 0.0 : color.r;
	color.g = color.g > 1.0 ? 0.0 : color.g;
	color.b = color.b > 1.0 ? 0.0 : color.b;

	return color;
}






vec2 EncodeNormal(vec3 normal)
{
	float p = sqrt(normal.z * 8.0 + 8.0);
	return vec2(normal.xy / p + 0.5);
}

vec3 DecodeNormal(vec2 enc)
{
	vec2 fenc = enc * 4.0 - 2.0;
	float f = dot(fenc, fenc);
	float g = sqrt(1.0 - f / 4.0);
	vec3 normal;
	normal.xy = fenc * g;
	normal.z = 1.0 - f / 2.0;
	return normal;
}






vec4 GetViewPosition(in vec2 coord, in float depth) 
{	
	vec2 tcoord = coord;
	TemporalJitterProjPosInv01(tcoord);

	vec4 fragposition = gbufferProjectionInverse * vec4(tcoord.s * 2.0f - 1.0f, tcoord.t * 2.0f - 1.0f, 2.0f * depth - 1.0f, 1.0f);
		 fragposition /= fragposition.w;
	
	return fragposition;
}

vec4 GetViewPositionNoJitter(in vec2 coord, in float depth) 
{	
	vec4 tcoord = vec4(coord.xy, 0.0, 0.0);

	vec4 fragposition = gbufferProjectionInverse * vec4(tcoord.s * 2.0f - 1.0f, tcoord.t * 2.0f - 1.0f, 2.0f * depth - 1.0f, 1.0f);
		 fragposition /= fragposition.w;

	
	return fragposition;
}

vec3 ProjectBack(vec3 cameraSpace) 
{
    vec4 clipSpace = gbufferProjection * vec4(cameraSpace, 1.0);
    vec3 NDCSpace = clipSpace.xyz / clipSpace.w;
    vec3 screenSpace = 0.5 * NDCSpace + 0.5;
		 //screenSpace.z = 0.1f;
    return screenSpace;
}

float 	ExpToLinearDepth(in float depth)
{
	#if 0
		return 2.0f * near * far / (far + near - (2.0f * depth - 1.0f) * (far - near));
	#else

		#if 1

		vec2 a = vec2(depth * 2.0 - 1.0, 1.0);

		// vec2 d = vec2(dot(vec2(gbufferProjectionInverse[2].z, gbufferProjectionInverse[3].z), a),
		// 			  dot(vec2(gbufferProjectionInverse[2].w, gbufferProjectionInverse[3].w), a));

		mat2 ipm = mat2(gbufferProjectionInverse[2].z, gbufferProjectionInverse[2].w,
						gbufferProjectionInverse[3].z, gbufferProjectionInverse[3].w);

		vec2 d = ipm * a;

		d.x /= d.y;

		return -d.x;

		#else

		vec4 d = gbufferProjectionInverse * vec4(0, 0, depth * 2.0 - 1.0, 1.0);
		d /= d.w;

		return -d.z;
		#endif

	#endif
}

float GetDepthLinear(in vec2 coord) 
{					
	return (near * far) / (texture2D(depthtex1, coord).x * (near - far) + far);
}

float GetDepth(vec2 coord)
{
	return texture2D(depthtex1, coord).x;
}

vec3 GetNormals(vec2 coord)
{
	return DecodeNormal(texture2D(colortex2, coord).xy);
}









vec4 SampleLinear(sampler2D tex, vec2 coord)
{
	return pow(texture2D(tex, coord), vec4(2.2));
}

vec3 LinearToGamma(vec3 c)
{
	return pow(c, vec3(1.0 / 2.2));
}

vec3 GammaToLinear(vec3 c)
{
	return pow(c, vec3(2.2));
}





float curve(float x)
{
	return x * x * (3.0 - 2.0 * x);
}

float Luminance(in vec3 color)
{
	return dot(color.rgb, vec3(0.2125f, 0.7154f, 0.0721f));
}









vec3 rand(vec2 coord)
{
	float noiseX = saturate(fract(sin(dot(coord, vec2(12.9898, 78.223))) * 43758.5453));
	float noiseY = saturate(fract(sin(dot(coord, vec2(12.9898, 78.223)*2.0)) * 43758.5453));
	float noiseZ = saturate(fract(sin(dot(coord, vec2(12.9898, 78.223)*3.0)) * 43758.5453));

	return vec3(noiseX, noiseY, noiseZ);
}

vec3 BlueNoiseTemporal(vec2 coord)
{
	vec2 noiseCoord = vec2(coord.st * vec2(viewWidth, viewHeight)) / 64.0;
	//noiseCoord += vec2(frameCounter, frameCounter);
	//noiseCoord += mod(frameCounter, 16.0) / 16.0;
	//noiseCoord += rand(vec2(mod(frameCounter, 16.0) / 16.0, mod(frameCounter, 16.0) / 16.0) + 0.5).xy;
	noiseCoord += vec2(sin(frameCounter * 0.75), cos(frameCounter * 0.75));

	noiseCoord = (floor(noiseCoord * 64.0) + 0.5) / 64.0;

	return texture2D(noisetex, noiseCoord).rgb;
}

float Get2DNoise(in vec2 pos)
{
	pos.xy += 0.5f;

	vec2 p = floor(pos);
	vec2 f = fract(pos);

	f.x = f.x * f.x * (3.0f - 2.0f * f.x);
	f.y = f.y * f.y * (3.0f - 2.0f * f.y);

	vec2 uv =  p.xy + f.xy;

	// uv -= 0.5f;
	// uv2 -= 0.5f;

	vec2 coord =  (uv  + 0.5f) / 64.0;
	float xy1 = texture2D(noisetex, coord).x;
	return xy1;
}

float Get3DNoise(in vec3 pos)
{
	pos.z += 0.0f;

	pos.xyz += 0.5f;

	vec3 p = floor(pos);
	vec3 f = fract(pos);

	// f.x = f.x * f.x * (3.0f - 2.0f * f.x);
	// f.y = f.y * f.y * (3.0f - 2.0f * f.y);
	// f.z = f.z * f.z * (3.0f - 2.0f * f.z);

	vec2 uv =  (p.xy + p.z * vec2(17.0f)) + f.xy;
	vec2 uv2 = (p.xy + (p.z + 1.0f) * vec2(17.0f)) + f.xy;

	// uv -= 0.5f;
	// uv2 -= 0.5f;

	vec2 coord =  (uv  + 0.5f) / 64.0;
	vec2 coord2 = (uv2 + 0.5f) / 64.0;
	float xy1 = texture2D(noisetex, coord).x;
	float xy2 = texture2D(noisetex, coord2).x;
	return mix(xy1, xy2, f.z);
}



vec4 ToSH(float value, vec3 dir)
{
	const float PI = 3.14159265359;
	const float N1 = sqrt(4 * PI / 3);
	const float transferl1 = (sqrt(PI) / 3.0) * N1;
	const float transferl0 = PI;

	const float sqrt1OverPI = sqrt(1.0 / PI);
	const float sqrt3OverPI = sqrt(3.0 / PI);

	vec4 coeffs;

	coeffs.x = 0.5 * sqrt1OverPI * value * transferl0;
	coeffs.y = -0.5 * sqrt3OverPI * dir.y * value * transferl1;
	coeffs.z = 0.5 * sqrt3OverPI * dir.z * value * transferl1;
	coeffs.w = -0.5 * sqrt3OverPI * dir.x * value * transferl1; //TODO: Vectorize the math so it's faster

	return coeffs;
}


vec3 FromSH(vec4 cR, vec4 cG, vec4 cB, vec3 lightDir)
{
	const float PI = 3.14159265;

	const float N1 = sqrt(4 * PI / 3);
	const float transferl1 = (sqrt(PI) / 3.0) * N1;
	const float transferl0 = PI;

	const float sqrt1OverPI = sqrt(1.0 / PI);
	const float sqrt3OverPI = sqrt(3.0 / PI);

	vec4 sh;

	sh.x = 0.5 * sqrt1OverPI;
	sh.y = -0.5 * sqrt3OverPI * lightDir.y;
	sh.z = 0.5 * sqrt3OverPI * lightDir.z;
	sh.w = -0.5 * sqrt3OverPI * lightDir.x;

	vec3 result;
	result.r = sh.x * cR.x;
	result.r += sh.y * cR.y;
	result.r += sh.z * cR.z;
	result.r += sh.w * cR.w;

	result.g = sh.x * cG.x;
	result.g += sh.y * cG.y;
	result.g += sh.z * cG.z;
	result.g += sh.w * cG.w;

	result.b = sh.x * cB.x;
	result.b += sh.y * cB.y;
	result.b += sh.z * cB.z;
	result.b += sh.w * cB.w;

	return result.rgb;
}



vec3 DoNightEyeAtNight(in vec3 color, float timeFactor)
{
	float luminance = Luminance(color);

	float rodFactor = 1.0f / (luminance * 1000.0 + 1.0);

	color = mix(color, luminance * vec3(0.2, 0.4, 0.9) * 1.5, vec3(0.6 * rodFactor * timeFactor));

	return color;
}












#if 0


//x is distance to outer surface, y is distance to inner surface
vec2 RaySphereIntersection( vec3 p, vec3 dir, float r ) 
{
	float b = dot( p, dir );
	float c = dot( p, p ) - r * r;
	
	float d = b * b - c;
	if ( d < 0.0 ) 
	{
		return vec2( 10000.0, -10000.0 );
	}

	d = sqrt( d );
	
	return vec2( -b - d, -b + d );
}


#define R_INNER 0.985

// Mie
// g : ( -0.75, -0.999 )
//      3 * ( 1 - g^2 )               1 + c^2
// F = ----------------- * -------------------------------
//      2 * ( 2 + g^2 )     ( 1 + g^2 - 2 * g * c )^(3/2)
float phase_mie( float g, float c, float cc ) {
	float gg = g * g;
	
	float a = ( 1.0 - gg ) * ( 1.0 + cc );

	float b = 1.0 + gg - 2.0 * g * c;
	b *= sqrt( b );
	b *= 2.0 + gg;	
	
	return 1.5 * a / b;
}

// Reyleigh
// g : 0
// F = 3/4 * ( 1 + c^2 )
float phase_reyleigh( float cc ) 
{
	return 0.75 * ( 1.0 + cc );
}

float density( vec3 p )
{
	const float R = 1.0;
	const float SCALE_H = 4.0 / ( R - R_INNER );
	const float SCALE_L = 1.0 / ( R - R_INNER );

	return exp( -( length( p ) - R_INNER ) * SCALE_H ) * 2.0;
}

float optic( vec3 p, vec3 q ) 
{
	const int numOutscatter = 1;

	const float R = 1.0;
	const float SCALE_L = 1.0 / (R - R_INNER);

	vec3 step = ( q - p ) / float(numOutscatter);
	step *= 0.3;
	vec3 v = p + step * 0.5;
	
	float sum = 0.0;
	for ( int i = 0; i < numOutscatter; i++ ) 
	{
		sum += density( v );
		v += step;
	}
	sum *= length( step ) * SCALE_L;


	return sum;
}

vec3 in_scatter(vec3 rayOrigin, vec3 rayDir, vec2 originToSkyIntersection, vec3 lightVector, float mieAmount, float rayleighAmount) 
{
	const float numInscatter = 3;
	
	const float PI = 3.14159265359;

	const float R = 1.0;
	const float SCALE_L = 1.0 / (R - R_INNER);

	float K_R = 0.186 * rayleighAmount;
	float K_M = 0.015 * mieAmount;
	const float E = 34.0;
	const vec3 C_R = vec3(0.4, 0.6, 1.0);	//Rayleigh scattering coefficients
	const float G_M = -0.82;


	float rayStepSize = (originToSkyIntersection.y * (1.0)) / float(numInscatter);
	vec3 step = rayDir * rayStepSize;
	step *= 2.0;

	//float boosty = 1.0 - abs(l.y);
	

	vec3 skyRayPos = rayOrigin + rayDir * ( rayStepSize * (0.5) );



	vec3 sum = vec3( 0.0 );
	for ( int i = 0; i < numInscatter; i++ ) 
	{
		vec2 atmosphereIntersection = RaySphereIntersection( skyRayPos, lightVector, R );
		vec3 outerAtmospherePos = skyRayPos + lightVector * atmosphereIntersection.y;
		
		float n = ( optic( rayOrigin, skyRayPos ) + optic( skyRayPos, outerAtmospherePos ) ) * ( PI * 4.0 );
		
		sum += density( skyRayPos ) * exp( -n * ( K_R * C_R ) );

		skyRayPos += step;
	}
	sum *= rayStepSize * SCALE_L;
	
	float c  = dot( rayDir, -lightVector );
	float cc = c * c;
	
	return sum * ( K_R * C_R * phase_reyleigh( cc ) + K_M * phase_mie( G_M, c, cc ) ) * E;
}



vec3 AtmosphericScattering(vec3 rayDir, vec3 lightVector, float mieAmount)
{
	const float R = 1.0;

	vec3 eye = vec3(0.0, mix(R_INNER, 1.0, 0.05), 0.0);

	vec3 originalRayDir = rayDir;

	if (rayDir.y < 0.0)
	{
		//rayDir.y = abs(rayDir.y);
		//rayDir.y *= rayDir.y;
		rayDir.y = 0.0;
		// rayDir.y *= -0.5;
	}
	rayDir.y = mix(rayDir.y, 1.0, 0.03);

	vec2 e = RaySphereIntersection(eye, rayDir, R);


	vec3 atmosphere = in_scatter(eye, rayDir, e, lightVector, mieAmount, 1.0);

	// float secondaryLightFactor = lightVector.y < 0.0 ? lightVector.y * 1.5 : lightVector.y;
	float secondaryLightFactor = lightVector.y;
	float secondary = pow(clamp(secondaryLightFactor, -1.0, 0.3) * 0.5 + 0.5, 8.0);
	// secondary *= 20.0 / (rayDir.y * 3.0 + 1.2);
	secondary *= 10.0;
	vec3 ambient = vec3(0.3, 0.5, 1.0);
	atmosphere += secondary * ambient;


	atmosphere *= vec3(0.4, 0.605, 1.0);
	// atmosphere *= vec3(0.6, 0.75, 1.0);

	// atmosphere *= exp(-0.5 * vec3(0.2, 0.45, 1.0));

	return atmosphere;
}

vec3 ModulateSkyForRain(vec3 skyColor, vec3 colorSkylight, float rainStrength)
{
	skyColor = mix(skyColor, vec3(1.0) * Luminance(colorSkylight), vec3(rainStrength * 0.95));

	return skyColor;
}

vec3 AtmosphericScattering(vec3 rayDir, vec3 lightVector, float mieAmount, float depth)
{
	const float R = 1.0;

	vec3 eye = vec3(0.0, mix(R_INNER, 1.0, 0.05), 0.0);

	vec3 originalRayDir = rayDir;

	if (rayDir.y < 0.0)
	{
		//rayDir.y = abs(rayDir.y);
		//rayDir.y *= rayDir.y;
		rayDir.y = 0.0;
	}

	vec3 up = vec3(0.0, 1.0, 0.0);

	vec2 e = RaySphereIntersection(eye, rayDir, R);
	e.y = depth;


	vec3 atmosphere = in_scatter(eye, rayDir, e, lightVector, mieAmount, 1.0);

	// float secondaryLightFactor = lightVector.y < 0.0 ? lightVector.y * 1.5 : lightVector.y;
	float secondaryLightFactor = lightVector.y;
	float secondary = pow(clamp(secondaryLightFactor, -1.0, 0.3) * 0.5 + 0.5, 8.0) * depth;
	secondary *= 10.0 / (rayDir.y * 3.0 + 1.2);
	vec3 ambient = vec3(0.3, 0.5, 1.0);
	atmosphere += secondary * ambient;


	atmosphere *= vec3(0.4, 0.605, 1.0);
	// atmosphere *= vec3(0.6, 0.75, 1.0);

	// atmosphere *= exp(-0.5 * vec3(0.2, 0.45, 1.0));






	return atmosphere;
}

vec3 AtmosphericScatteringSingle(vec3 rayDir, vec3 lightVector, float mieAmount)
{
	const float PI = 3.14159265359;
	const float DEG_TO_RAD = PI / 180.0;

	//Scatter constants
	const float K_R = 0.166;
	const float K_M = 0.0025;
	const float E = 14.3;
	const vec3 C_R = vec3(0.3, 0.7, 1.0);	//Rayleigh scattering coefficients
	const float G_M = -0.85;

	const float R = 1.0;
	const float SCALE_H = 4.0 / (R - R_INNER);
	const float SCALE_L = 1.0 / (R - R_INNER);

	const int NUM_OUT_SCATTER = 10;
	const float FNUM_OUT_SCATTER = 10.0;

	const int NUM_IN_SCATTER = 10;
	const float FNUM_IN_SCATTER = 10.0;

	vec3 eye = vec3(0.0, mix(R_INNER, 1.0, 0.05), 0.0);

	vec3 originalRayDir = rayDir;

	if (rayDir.y < 0.0)
	{
		//rayDir.y = abs(rayDir.y);
		//rayDir.y *= rayDir.y;
		rayDir.y = 0.0;
	}

	vec3 up = vec3(0.0, 1.0, 0.0);

	vec2 e = RaySphereIntersection(eye, rayDir, R);
	vec2 eup = RaySphereIntersection(eye, up, R);


	vec3 atmosphere = in_scatter(eye, rayDir, e, lightVector, mieAmount, 0.7);


	// atmosphere = pow(atmosphere, vec3(1.2));
	atmosphere = pow(atmosphere, vec3(1.9));

	//if (originalRayDir.y < 0.0)
	//{
		//atmosphere *= curve(saturate(originalRayDir.y + 1.0));
	//}


	return atmosphere;
}


#else




// Mie
// g : ( -0.75, -0.999 )
//      3 * ( 1 - g^2 )               1 + c^2
// F = ----------------- * -------------------------------
//      2 * ( 2 + g^2 )     ( 1 + g^2 - 2 * g * c )^(3/2)
float PhaseMie( float g, float LdotV, float LdotV2 ) {
	float gg = g * g;
	
	float a = ( 1.0 - gg ) * ( 1.0 + LdotV2 );

	float b = 1.0 + gg - 2.0 * g * LdotV;
	b *= sqrt( b );
	b *= 2.0 + gg;	
	
	return 1.5 * a / b;
}

// Reyleigh
// g : 0
// F = 3/4 * ( 1 + c^2 )
float PhaseRayleigh( float LdotV2 ) 
{
	return 0.75 * ( 1.0 + LdotV2 );
}


float TotalAtmosphereDensity(float dirY)
{
	// return 1.0 / (max(0.0, dirY + 0.06) * 3.0 + 0.00001);
	return 1.0 / (SmoothMax(0.0, dirY + 0.03, 0.1) * 3.0 + 0.00001);
}

const vec3 Rayleigh = pow(vec3(0.17, 0.44, 0.99), vec3(1.2));
// const vec3 Rayleigh = vec3( 3.8, 13.5, 33.1 ) * 0.05;

vec3 SunlightColorFromSunVector(vec3 worldSunVector)
{

	float thickness = TotalAtmosphereDensity(worldSunVector.y);

	vec3 color = exp(-thickness * 1.0 * Rayleigh);

	color *= pow(saturate(worldSunVector.y), 0.5);

	return color;
}

vec3 Atmosphere(vec3 worldViewVector, vec3 worldSunVector, float mieAmount)
{
	// worldViewVector.y = abs(worldViewVector.y) * 0.99 + 0.01;
	worldViewVector.y = SmoothMax(0.0, worldViewVector.y, 0.1);
	// worldViewVector.y = max(0.0, worldViewVector.y);

	// return vec3(1.0);

	float LdotV = dot(worldViewVector, worldSunVector);
	float LdotV2 = LdotV * LdotV;

	vec3 localRayleigh = pow(Rayleigh, vec3(saturate(worldViewVector.y * 0.5 + 0.5)));
	// vec3 localRayleigh = Rayleigh;


	vec3 color = vec3(0.0, 0.0, 0.0);


	// float sunThroughAtmosphereDensity = pow(saturate(dot(normalize(worldViewVector + vec3(0.0, 55.0 * worldViewVector.y, 0.0)), worldSunVector) * 0.6 + 0.4), 1.0);
	float sunThroughAtmosphereDensity = TotalAtmosphereDensity(worldSunVector.y);


	// float thickness = 1.0 / (saturate(worldViewVector.y) * 5.0 + 0.2);
	float thickness = 1.0 / ((worldViewVector.y) * (1.0 + sunThroughAtmosphereDensity * 4.0) + 0.25);
	// float rawDensity = 1.0 / ((worldViewVector.y) * 2.0 + 0.15);
	float density = pow(thickness, 1.2);
	thickness *= sunThroughAtmosphereDensity * 0.75;
	thickness *= 1.0 + (-(LdotV) * 0.5 + 0.5) * 0.025 / (exp(-sunThroughAtmosphereDensity) + 0.01);

	float phaseRayleigh = PhaseRayleigh(LdotV2);
	color = exp(-thickness * localRayleigh) * localRayleigh * density * phaseRayleigh;
	// color += pow(localRayleigh, vec3(1.0)) * 0.38 * exp(-sunThroughAtmosphereDensity * 0.3) * (1.0 - exp(-rawDensity * 0.45));
	// color += localRayleigh * 0.7768 * exp(-sunThroughAtmosphereDensity * 0.3 ) * (1.0 - exp(-rawDensity * 0.45))
			 // * exp(-sunThroughAtmosphereDensity * localRayleigh * 0.35);

	// color += localRayleigh * 0.4768 * exp(-sunThroughAtmosphereDensity * 0.3 )
	// 	 * exp(-sunThroughAtmosphereDensity * localRayleigh * 0.35);	 

	color += localRayleigh * 0.268 * exp(-TotalAtmosphereDensity(worldSunVector.y + 0.02) * 0.3 );
		 // * exp(-sunThroughAtmosphereDensity * localRayleigh * 0.35);	


	float phaseMie = PhaseMie(0.4, LdotV, LdotV2);
	float phaseMie2 = PhaseMie(0.78, LdotV, LdotV2);
	color += phaseMie * 0.185 
	// * exp(-localRayleigh * TotalAtmosphereDensity(worldViewVector.y) * 0.15)
	// * exp(-localRayleigh * sunThroughAtmosphereDensity * 0.15)
	* exp(-thickness * localRayleigh)
	* density;

	color += phaseMie2 * 0.136 * exp(-localRayleigh * TotalAtmosphereDensity(worldViewVector.y) * sunThroughAtmosphereDensity * 0.3) * density;




	color *= 4.0;


	// color = vec3(0.1, 0.3, 1.0);

	// color *= SunlightColorFromSunVector(worldSunVector);

	return color;
}


vec3 AtmosphereAbsorption(vec3 worldViewVector)
{
	worldViewVector.y = max(worldViewVector.y, 0.0);

	float thickness = 1.0 / ((worldViewVector.y) * (1.0) + 0.15);
	float thicknessHaze = 1.0 / ((worldViewVector.y) * (10.0));

	return exp(-thickness * (Rayleigh)) * exp(-thicknessHaze);
}


// Main sky shading function
vec3 SkyShading(vec3 worldViewVector, vec3 worldSunVector, float rainStrength)
{
	vec3 atmosphere = Atmosphere(worldViewVector, worldSunVector, 1.0);
	atmosphere += Atmosphere(worldViewVector, -worldSunVector, 1.0) * NIGHT_BRIGHTNESS;

	// rainy
	vec3 rainSky = vec3(0.5) * (exp(-TotalAtmosphereDensity(worldSunVector.y)) + 1.0 * NIGHT_BRIGHTNESS);
	atmosphere = mix(atmosphere, rainSky, vec3(rainStrength));

	// return vec3(1.0);
	return atmosphere;
}

vec3 AtmosphericScattering(vec3 rayDir, vec3 lightVector, float mieAmount)
{
	return Atmosphere(rayDir, lightVector, mieAmount);
}

vec3 AtmosphericScattering(vec3 rayDir, vec3 lightVector, float mieAmount, float depth)
{
	return vec3(0.0);
}

vec3 AtmosphericScatteringSingle(vec3 rayDir, vec3 lightVector, float mieAmount)
{
	return AtmosphereAbsorption(rayDir);
}

vec3 ModulateSkyForRain(vec3 skyColor, vec3 colorSkylight, float rainStrength)
{
	skyColor = mix(skyColor, vec3(1.0) * Luminance(colorSkylight), vec3(rainStrength * 0.95));

	return skyColor;
}

float RenderSunDisc(vec3 worldDir, vec3 sunDir)
{
	float d = dot(worldDir, sunDir);

	float disc = 0.0;

	//if (d > 0.99)
	//	disc = 1.0;

	float size = 0.00195;
	float hardness = 1000.0;

	disc = pow(curve(saturate((d - (1.0 - size)) * hardness)), 2.0);

	float visibility = curve(saturate(worldDir.y * 30.0));

	disc *= visibility;

	return disc;
}




#endif







// Lighting colors and data
vec3 GetColorSunlight(vec3 worldSunVector, float rainStrength)
{
	vec3 color = SunlightColorFromSunVector(worldSunVector);

	color *= 1.0 - rainStrength;

	color += SunlightColorFromSunVector(-worldSunVector) * NIGHT_BRIGHTNESS;

	// return vec3(1.0);
	return color;
}

vec3 GetColorTorchlight()
{
	vec3 colorTorchlight = vec3(0.0);

	if (TORCHLIGHT_COLOR_TEMPERATURE == 2000)
		//2000k
		colorTorchlight = pow(vec3(255, 141, 11) / 255.0, vec3(2.2));
	else if (TORCHLIGHT_COLOR_TEMPERATURE == 2300)
		//2300k
		colorTorchlight = pow(vec3(255, 152, 54) / 255.0, vec3(2.2));
	else if (TORCHLIGHT_COLOR_TEMPERATURE == 2500)
		//2500k
		colorTorchlight = pow(vec3(255, 166, 69) / 255.0, vec3(2.2));
	else
		//3000k
		colorTorchlight = pow(vec3(255, 180, 107) / 255.0, vec3(2.2));

	return colorTorchlight;
}

void GetSkylightData(vec3 worldSunVector, float rainStrength,
	out vec4 skySHR, out vec4 skySHG, out vec4 skySHB, 
	out vec3 colorSkylight, out vec3 colorSkyUp)
{
	colorSkylight = vec3(0.0);

	const int latSamples = 5;
	const int lonSamples = 5;

	skySHR = vec4(0.0);
	skySHG = vec4(0.0);
	skySHB = vec4(0.0);

	for (int i = 0; i < latSamples; i++)
	{
		float latitude = (float(i) / float(latSamples)) * 3.14159265;
			  latitude = latitude;
		for (int j = 0; j < lonSamples; j++)
		{
			float longitude = (float(j) / float(lonSamples)) * 3.14159265 * 2.0;

			vec3 rayDir;
			rayDir.x = cos(latitude) * cos(longitude);
			rayDir.z = cos(latitude) * sin(longitude);
			rayDir.y = sin(latitude);

			vec3 skyCol = SkyShading(rayDir, worldSunVector, rainStrength);
			colorSkylight += skyCol;

			skySHR += ToSH(skyCol.r, rayDir);
			skySHG += ToSH(skyCol.g, rayDir);
			skySHB += ToSH(skyCol.b, rayDir);
		}
	}

	skySHR /= latSamples * lonSamples;
	skySHG /= latSamples * lonSamples;
	skySHB /= latSamples * lonSamples;

	colorSkylight /= latSamples * lonSamples;
	colorSkyUp = SkyShading(vec3(0.0, 1.0, 0.0), worldSunVector, rainStrength);
}





// Time data
float GetTimeMidnight(vec3 worldSunVector)
{
	return saturate(-worldSunVector.y * 10.0);
}